# -*- coding: utf-8 -*-
#
# Copyright 2022 Nitrokey Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

from typing import Iterable, Optional

class tqdm:
    def __init__(
        self,
        iter: Optional[Iterable] = None,
        total: Optional[int] = None,
        unit: Optional[str] = None,
        unit_scale: Optional[bool] = None,
        leave: Optional[bool] = None,
    ) -> None: ...
    def update(self, n: int) -> None: ...
    def close(self) -> None: ...
    def __iter__(self): ...
